#include <csignal>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <algorithm>
#include <unistd.h>

#include <getopt.h>

#include <string>

using namespace std;

#include "serverfun.h"
#include "resample.hh"
#include "filters.h"

static unsigned Filter = 1;  /* linear */
static double blur = 1;      /* <1 = sharp, >1 = blurry */
static bool dbl1=false, dbl2=false, flip1=false, flip2=false;

static void CheckFlip(stent *Buf, unsigned count)
{
   	if(dbl1) for(unsigned a=0; a<count; ++a) Buf[a].l += Buf[a].l;
   	if(dbl2) for(unsigned a=0; a<count; ++a) Buf[a].r += Buf[a].r;
   	if(flip1) for(unsigned a=0; a<count; ++a) Buf[a].l = -Buf[a].l;
   	if(flip2) for(unsigned a=0; a<count; ++a) Buf[a].r = -Buf[a].r;
}

/* This the function which started the program development.
 * The core of this program. The rest is just waveform
 * conversions and parameter handling...
 */
static void Interpolate(Stream &input, Stream &output)
{
    if(input.rate() == output.rate())
    {
        // No change in rate, just copy
        stent tab[512];
        for(;;)
        {
            unsigned oco = input.Read(tab, 512);
            if(!oco)return;
            CheckFlip(tab, 512);
            output.Write(tab, 512);
        }
    }
    
    double factor = (double)input.rate() / output.rate();
    
    double scale_factor = blur * max(factor, 1.0);
    double support = max(scale_factor * Filters[Filter].support, 0.5);
    if(support <= 0.5)
    {
        /* reduce to point sampling (??) */
        support = 0.5 + 1E-12;
        scale_factor = 1.0;
    }
    
    double *weight = (double *)alloca((sizeof(double)) * ((int)support*2+3));
    double (*const filterfun)(double) = Filters[Filter].function;
    
    double center = 0.0;
    const unsigned outbufsize = output.rate() / 8;
    
    stent *oent = (stent *)alloca(outbufsize * sizeof(stent));
    
    /* This resampling algorithm is based on ideas from ImageMagick. */
    for(;;)
    {
        for(unsigned o=0; o<outbufsize; ++o, center += factor)
        {
            int start = (int)(0.5 + center - support);
            int end   = (int)(0.5 + center + support);
            
            register double density = 0;
            int n = 0;
            for(int i = start; i < end; ++i)
                density += (weight[n++] = filterfun((i-center+0.5) / scale_factor) / scale_factor);
            
            if(density != 0.0)
                for(register int i=0; i<n; ++i)
                    weight[i] /= density; /* normalize */
            
            register double l=0.0;
            register double r=0.0;
            for(register int i=0, p=start; i<n; ++i)
            {
                const stent &srcent = input[p++];
                l += weight[i] * srcent.l;
                r += weight[i] * srcent.r;
            }
            oent[o].l = l<-32768 ? -32768 : l>32767 ? 32767 : (int)(l+0.5);
            oent[o].r = r<-32768 ? -32768 : r>32767 ? 32767 : (int)(r+0.5);
        }

        output.Write(oent, outbufsize);
        output.Flush();
        
        register int shiftval = (int)center;
        /* Prevent overflowing in 17 hours */
        input.shift(shiftval);
        center -= shiftval;
    }
}

Stream::Stream(const char *settings) : bit16(true), stereo(true),
                                       herz(0), basepos(-inbufsize)
{
    while(*settings)
    {
        if(*settings=='m')stereo = false;
        else if(*settings=='s')stereo = true;
        else if(*settings=='b')bit16 = false;
        else if(*settings=='w')bit16 = true;
        else if(*settings=='r')herz = 0;
        else if(*settings>='0' && *settings<='9')
            herz = herz * 10 + (*settings - '0');
        else if(*settings!='-')throw *settings;
        ++settings;
    }
    if(!herz)herz = 44100;
}

const stent &Stream::operator[] (int index)
{
    while(index >= basepos + inbufsize)
    {
        basepos += leapstep;
        memmove(&inent[0], &inent[leapstep], min(leapstep,newsize) * sizeof(inent[0]));
        int rval = Read(&inent[leapstep], newsize);
        if(!rval)
        {
            /* EOF */
            exit(0);
        }
        if(rval < newsize)
            memset(&inent[leapstep+rval], 0, (newsize-rval)*sizeof(inent[0]));
    }
    if(index < basepos)return inent[0];
    return inent[index - basepos];
}

unsigned Stream::Read(stent *dest, unsigned count) const
{
    if(bit16)
        if(stereo)
            return SafeRead(dest, count*4) / 4;
        else
            return read16(dest, count);
    else
        if(stereo)
            return readste(dest, count);
        else
            return read8(dest, count);
	CheckFlip(dest, count);
}
void Stream::Write(const stent *dest, unsigned count) const
{
    if(bit16)
        if(stereo)
            SafeWrite(dest, count*4);
        else
            write16(dest, count);
    else
        if(stereo)
            writeste(dest, count);
        else
            write8(dest, count);
}

void Stream::Flush() const
{
    fdatasync(1);
}

/* Read 16-bit mono entries */
unsigned Stream::read16(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    count = SafeRead(tab, bytes) / 2;
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a];
    }
    return count;
}
/* Read 8-bit stereo entries */
unsigned Stream::readste(stent *dest, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead(tab, bytes) / 2;
    for(unsigned a=0, b=0; b<count; ++b)
    {
        dest[b].r = tab[a++]*256 - 32768;
        dest[b].l = tab[a++]*256 - 32768;
    }
    return count;
}
/* Read 8-bit mono entries */
unsigned Stream::read8(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead(tab, bytes);
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a]*256 - 32768;
    }
    return count;
}

void Stream::write16(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 16-bit stereo to mono mixing */
        tab[a] = (src[a].r + src[a].l) / 2;
    }
    SafeWrite(tab, bytes);
}
void Stream::writeste(const stent *src, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0, b=0; b<count; ++b)
    {
        tab[a++] = (src[b].r + 32768) / 256;
        tab[a++] = (src[b].l + 32768) / 256;
    }
    SafeWrite(tab, bytes);
    
}
void Stream::write8(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 8-bit stereo to mono mixing */
        tab[a] = ((src[a].r + src[a].l) / 2 + 32768) / 256;
    }
    SafeWrite(tab, bytes);
}

#define VERSIONSTR \
    "resample v"VERSION" - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "Portions copyright (C) 2001 ImageMagick Studio.\n" \
    "This program is under GPL. streamixer-"VERSION".tar.{gz,bz2}\n" \
    "are available at the homepage of the author.\n"

static void Killer(int)
{
    exit(-1);
}

int main(int argc, char** argv)
{
    signal(SIGPIPE, Killer);

    Stream *input = NULL, *output = NULL;
    bool Heelp = false;
    bool stderr_helper = false;
    
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0, 0,'h'},
            {"version",  0, 0,'V'},
            {"blur",     1, 0,'b'},
            {"lflip",    0, 0,'L'},
            {"rflip",    0, 0,'R'},
            {"ldouble",  0, 0,502},
            {"rdouble",  0, 0,503},
            {"filter",   1, 0,'f'},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "HhVb:f:", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'H': stderr_helper = true; break;
            case 'h': Heelp = true; break;
            case 'V':
                printf("%s", VERSIONSTR);
                return 0;
            case 'b': blur = 1 + strtol(optarg,NULL,10); break;
            case 's': blur = 1 - strtol(optarg,NULL,10); break;
            case 'L': flip1 = true; break;
            case 'R': flip2 = true; break;
            case 502: dbl1 = true; break;
            case 503: dbl2 = true; break;
            case 'f':
            {
                std::string s = optarg;
                unsigned a;
                for(a=0; a<FilterCount; ++a)
                    if(Filters[a].name == s)
                    {
                        Filter = a;
                        break;
                    }
                if(a < FilterCount)break;
                fprintf(stderr, "'%s' is not a supported filter.\n", s.c_str());
                return -1;
            }
            default:
                break;
        }
    }
    
    if(argc != optind+2)
    {
        Heelp = true;
    }
    
    if(Heelp)
    {
        fprintf(stderr,
            "%s\n"
            "Usage: %s <inputspec> <outputspec> [<options> ...]\n"
            "\n"
            "Reads stdin, writes stdout.\n"
            "\n"
            "input- and outputspec:\n"
            "   b=8-bit, m=mono, rxxx=xxx Hz, examples:\n"
          "       \"\"      = 16-bit 44100 Hz stereo\n"
            "       m       = 16-bit 44100 Hz mono\n"
            "       b       = 8-bit 44100 Hz stereo\n"
            "       mb      = 8-bit 44100 Hz mono\n"
            "       mr22050 = 16-bit 22050 Hz mono\n"
            "       br22050 = 8-bit 22050 Hz stereo\n"
            "   etc. '-' must not be used.\n"
            "\nOptions:\n"
            "  --help, -h              Help\n"
            "  --version, -V           Version information\n"
            "  --blur, -b <amount>     Blur: 0=normal, >0=blur (cpu intensive)\n"
            "  --lflip, -L             Flip left channel\n"
            "  --rflip, -R             Flip right channel\n"
            "  --ldouble               Multiply left channel amplitude by 2\n"
            "  --rdouble               Multiply right channel amplitude by 2\n"
            "  --sharpen, -s <amount>  Equivalent to giving negative number to --blur\n"
            "  --filter, -f <filter>   Select filter (interpolation scheme)\n"
            "\n"
            "Available filters:\n"
            "  none, linear, triangle, hermite, hanning, hamming,\n"
            "  blackman, gaussian, quadratic, cubic, catrom, mitchell,\n"
            "  lanczos, bessel, sinc\n"
            "Recommended filters: linear, lanczos, none\n"
            "\n",
            VERSIONSTR, argv[0]);

        return 0;
    }

    try
    {
        input = new Stream(argv[optind+0]);
        output = new Stream(argv[optind+1]);
    }
    catch(char c)
    {
        fprintf(stderr, "Invalid option: '%c' (-switches must be specified after stream parameters)\n", c);
        return -1;
    }
    
        
    if(stderr_helper)
    {
        static const char c = '!';
        write(2, &c, 1);
    }
    Interpolate(*input, *output);
    return 0;
}
